
/*
	 Program generator liczb losowych, 2002 I, v1.0,  GPL.
         Projekt CSI - Zlozonosc obliczeniowa, struktury danych i algorytmy
	 
	 
	 autor:
	 
	    Piotr Skowronek <csi22@m8.mech.pk.edu.pl> 
		            <skylark@wizard.ae.krakow.pl>
			    http://www.sky-lark.tk
			    http://wizard.ae.krakow.pl/~skylark
	 
	 Program stosuje generator kongruentny X(n+1)=a*X(n) (modM),
	 generujace liczby pseudo-losowe z przedzialu (0,1).
	 Program przeprowadza nastepujace testy generatorow:
	 - test Weila 
	 - test Pi
	 - test Chi^2 (X^2)
	 
	 Test Weila jest to test generatorow liczb losowych.
	 Polega on na wyliczeniu sredniej arytmetycznej z wylosowanych
	 liczb. Wartosc srednia powinna byc rowna 0.5 dla idealnego
	 generatora.
	 
	 Test Pi jest to test generatorow liczb losowych. 
	 Polega on  na losowaniu pary liczb z przedzialu (0, 1) i  
	 sprawdzeniu czy znajduje sie ona wewnatrz kola o promieniu 1. 
	 Jezeli tak jest to jest to sukces. Prawdopodobienstwo,   
	 ze punkt znajduje sie wewnatrz cwiartki kola jest rowne         
	 ilorazowi cwiertki pola kola przez pole kwadratu o boku 1 
	 czyli Pi/4 a to jest rowne ilosci sukcesow przez ilosc par.
	 Stad mozna obliczyc Pi, porownac z rzeczywista i powiedziec cos          
	 o generatorze.    
	 
	 Test Chi^2 zostal opisany na wykladzie.                                                 
	 
		     
	 Kompilacja:
	 g++ rand.cpp -o rand
	 Wywolanie:
	 ./rand  	 
	 
	 Program sam prowadzi uzytkownika.
*/


#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>


long iran=0, ia, im, iq, ir;
long R=1000000;			// ilosc mozliwych liczb jakie
				// generator moze wygenerowac


void sky_srand()		// funkcja inicjalizujaca generator
{
	time_t time_t;
	im=2147483647;		// Wartosci stalych dla generatora
	ia=69621;
	
	iq=im/ia;		// obliczenie paremtrow 
	ir=im%ia;		// potrzebnych do konwersji Schrage
	
	time(&time_t);
	iran=time_t%im-1;
};

float sky_rand()		// funkcja generujaca liczby losowe
{				// X(n+1)=a*X(n)(modM)	
	if (!iran) sky_srand();
	
	long ik=iran/iq;
	iran=ia*(iran-ik*iq)-ir*ik;
	if (iran<0) iran+=im;
	return (iran/(float) im);		
};


int main(int argc, char *argv[])
{
	long ilosc;
	float suma=0, sukces=0, x, y;
	char *ch2_tab;
	
	
	cout << "Demonstracja generatora liczb losowych wraz ze statystykami." << endl
	     << "Projekt CSI - Zlozonosc obliczeniowa, struktury danych i algorytmy
" << endl	 
	     << "Piotr Skowronek, csi22@m8.mech.pk.edu.pl, skylark@wizard.ae.krakow.pl" << endl
	     << "2001 XII. v1.0. GPL" << endl;

	do
	{
	cout << "Podaj ilosc liczb jakie mam wylosowac (>10*1000000): " << endl;
	cin >> ilosc;
	} while (ilosc<=0);
	
	if (!(ch2_tab=(char *)malloc(sizeof(char)*R)))
	{
	    cout << "BLAD: zamalo pamieci aby przeprowadzic test." << endl;
	    return -1;
	};
	
	memset(ch2_tab,0,sizeof(char)*R); // wyczyszczenie pamieci
	
	sky_srand();  		// Inicjalizacja generatora

	cout << "Generowanie " << ilosc << " liczb..." << endl;
	for (long i=0; i<ilosc/2; i++) 
	    { 
		x=sky_rand();	// losowanie pary liczb
		y=sky_rand();
		
		// ******** potrzebne do chi^2
		ch2_tab[(long long) (x*R)]+=1;  // zwiekszenie licznika 
		ch2_tab[(long long) (y*R)]+=1;  // wystapien danej liczby
		
		// ******** potrzebne do testu Weila
		suma+=x+y; 
		
		// ******** potrzebne do testu PI
				// sprawdzenie czy mieszcza sie w kole o r=(0,1)
		if (x*x+y*y<=1) sukces++;	
		
		// ******** informacja n/t postepow
		if (i%1000==0) cout << '\r' << (float) i/(ilosc/2) *100 << "%     ";
	    };
	cout << "\rZrobione" << endl << endl;
	
	
	//************* Test Sredniej arytmetycznej
	cout << "Test sredniej arytmetycznej (powinna byc bliska 0.5)." << endl
	     << "Sr=" << suma/ilosc << endl;
	     
	//************* Test Pi
	cout << "Test PI. Wartosc P powinna byc bliska Pi (3.1415926536). " << endl
	     << "P="  << (sukces/(ilosc/2))*4 << endl;

	//************* Statystyka Chi^2
        cout << "Obliczanie statystyki Chi^2..." << endl;	

	long double wynik=0;
	long double N_div_R=(float)ilosc/(float)R;  // obliczamy raz, i potem 
						    // korzystamy z tego w petli

	for (long i=0; i<R; i++)
	    wynik=wynik+(long double)(((ch2_tab[i]-N_div_R)*(ch2_tab[i]-N_div_R))/N_div_R);
	    
	cout << "Chi^2=" << (long long) wynik << endl;
	if ((wynik+2*sqrt(R)>=R) && (R>=wynik-2*sqrt(R)))
	    cout << "Generator przeszedl pomyslnie test." << endl;
	else
	    cout << "Generator nie przeszedl pomyslnie testu." << endl
		 << "Sprobuje ponownie, poniewaz 1 na 10 testow bywa nie poprawnych." << endl;
	
	return 0;
};